import json
import os
from typing import Optional


class FileHandler:
    def __init__(self, directory: str):
        self.cache: dict[str, dict] = {}
        self.directory: str = directory

    def get_files(self, base_path: str = ""):
        root_path: str = os.path.abspath(self.directory)
        full_path: str = os.path.join(root_path, base_path)
        full_path = os.path.abspath(full_path)
        common_prefix: str = os.path.commonprefix([full_path, root_path])

        if common_prefix != root_path:
            return []

        return [
            os.path.join(base_path, f)
            for f in os.listdir(full_path)
        ]

    def get_files_meta(self, base_path: str = ""):
        files: list[str] = self.get_files(base_path)
        files = [
            os.path.join(self.directory, f)
            for f in files
        ]
        files_meta: list[dict] = []

        deleted = set(self.cache.keys()) - set(files)
        for path in deleted:
            del self.cache[path]

        for path in files:
            last_modified: float = os.path.getmtime(path)
            if path not in self.cache or self.cache[path]["ts"] < last_modified:
                self.update_meta(path)

            files_meta.append(self.cache[path])

        return files_meta

    def update_meta(self, path: str) -> None:
        self.cache[path] = self.get_meta(path)

    def get_meta(self, path: str) -> dict:
        return {
            "path": os.path.relpath(path, self.directory),
            "filename": os.path.basename(path),
            "ts": os.path.getmtime(path)
        }


class JsonFileHandler(FileHandler):
    def read(self, path: str) -> Optional[dict|list]:
        if path not in self.get_files():
            return None
        with open(os.path.join(self.directory, path), "r") as f:
            data = json.load(f)
        return data

    def write(self, path: str, data: dict|list) -> bool:
        if path not in self.get_files():
            return False

        try:
            with open(os.path.join(self.directory, path), "w", encoding="utf-8") as f:
                json.dump(data, f, indent=2, ensure_ascii=False)
        except:
            return False
        return True


class MetadataFileHandler(JsonFileHandler):
    def get_meta(self, path: str) -> dict:
        meta: dict = super().get_meta(path)

        with open(path, "r") as f:
            data = json.load(f)
            is_series = "filename" not in data
            meta["type"] = "series" if is_series else "film"
            if is_series:
                meta["episodes"] = len(data)
                meta["title"] = meta["filename"].split("_metadata")[0]
            else:
                meta["title"] = data["title"]

        return meta


class ToConvertFileHandler(FileHandler):
    def get_meta(self, path: str) -> dict:
        meta: dict = super().get_meta(path)
        is_dir: bool = os.path.isdir(path)

        meta["size"] = os.path.getsize(path)
        meta["type"] = "folder" if is_dir else "media"
        if is_dir:
            meta["elements"] = len(os.listdir(path))
            if not meta["path"].endswith("/"):
                meta["path"] += "/"

        return meta
