import json
import logging
import os
import subprocess
from typing import Optional


class MetadataExtractor:
    SUPPORTED_EXTENSIONS = (".mp4", ".mkv", ".mov", ".avi")

    def __init__(self):
        self.logger: logging.Logger = logging.getLogger("MetadataExtractor")

    def analyze_file(self, path: str) -> Optional[dict]:
        """
        Extracts metadata from a video file using ffprobe

        :param path: Path to the video file
        :return: Metadata information or ``None`` if an error occurred
        """

        # Get general file info in JSON format
        cmd: list[str] = [
            "ffprobe",
            "-v", "quiet",
            "-print_format", "json",
            "-show_format",
            "-show_streams",
            path
        ]

        try:
            result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            if result.returncode != 0:
                self.logger.error(f"Error processing {path}: {result.stderr}")
                return None

            data: dict = json.loads(result.stdout)

            # Extract filename and title
            filename: str = os.path.basename(path)
            title: str = data.get("format", {}).get("tags", {}).get("title", filename)

            # Initialize metadata structure
            metadata: dict = {
                "filename": filename,
                "title": title,
                "audio_tracks": [],
                "subtitle_tracks": []
            }

            # Process streams
            for stream in data.get("streams", []):
                codec_type = stream.get("codec_type")
                tags = stream.get("tags", {})
                disposition = stream.get("disposition", {})
                track = {
                    "index": stream.get("index"),
                    "language": tags.get("language", "und"),
                    "name": tags.get("title", ""),
                    "flags": {
                        "default": disposition.get("default", 0) == 1,
                        "original": disposition.get("original", 0) == 1,
                        "commentary": disposition.get("commentary", 0) == 1
                    }
                }

                if codec_type == "audio":
                    track |= {
                        "channels": stream.get("channels", 0)
                    }
                    track["flags"] |= {
                        "visual_impaired": disposition.get("visual_impaired", 0) == 1
                    }
                    metadata["audio_tracks"].append(track)

                elif codec_type == "subtitle":
                    track["flags"] |= {
                        "forced": disposition.get("forced", 0) == 1,
                        "hearing_impaired": disposition.get("hearing_impaired", 0) == 1
                    }
                    metadata["subtitle_tracks"].append(track)

                elif codec_type == "video":
                    pass

                elif codec_type == "button":
                    pass

                else:
                    self.logger.warning(f"Unknown track codec type '{codec_type}'")

            return metadata

        except Exception as e:
            self.logger.error(f"Error processing {path}: {str(e)}")
            return None

    def process_file(self, file_path: str, output_dir: str) -> bool:
        """
        Processes a single video file and writes metadata to a JSON file

        :param file_path: Path of the video file
        :param output_dir: Path of the directory where the output JSON file will be saved
        :return: True if successful, False otherwise
        """

        if not os.path.isfile(file_path):
            self.logger.error(f"File not found: {file_path}")
            return False

        if not file_path.lower().endswith(self.SUPPORTED_EXTENSIONS):
            self.logger.error(f"Unsupported file format: {file_path}")
            return False

        self.logger.debug(f"Extracting metadata from {os.path.basename(file_path)}")
        metadata: Optional[dict] = self.analyze_file(file_path)

        if metadata:
            # Generate output filename based on input file
            filename = os.path.basename(os.path.splitext(file_path)[0]) + "_metadata.json"

            if output_dir:
                # Ensure output directory exists
                os.makedirs(output_dir, exist_ok=True)
                output_path = os.path.join(output_dir, filename)
            else:
                # If no output directory specified, save in the same directory as the input file
                base_name = os.path.splitext(file_path)[0]
                output_path = f"{base_name}_metadata.json"

            # Write metadata to JSON file
            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(metadata, f, indent=2, ensure_ascii=False)

            self.logger.debug(f"Metadata saved to {output_path}")
            return True

        return False

    def process_directory(self, directory_path: str, output_dir: Optional[str] = None) -> bool:
        """
        Processes all video files in a directory and writes metadata to a JSON file

        :param directory_path: Path of the directory
        :param output_dir: Path of the directory where the output JSON file will be saved
        :return: True if successful, False otherwise
        """

        if not os.path.isdir(directory_path):
            self.logger.error(f"Directory not found: {directory_path}")
            return False

        all_metadata: dict[str, dict] = {}
        file_count: int = 0

        for root, _, files in os.walk(directory_path):
            for file in files:
                if file.lower().endswith(self.SUPPORTED_EXTENSIONS):
                    file_path: str = os.path.join(root, file)
                    self.logger.debug(f"Extracting metadata from {file}")
                    metadata: Optional[dict] = self.analyze_file(file_path)

                    if metadata:
                        # Use relative path as key
                        rel_path: str = os.path.relpath(file_path, directory_path)
                        all_metadata[rel_path] = metadata
                        file_count += 1

        if file_count == 0:
            self.logger.error(f"No supported video files found in {directory_path}")
            return False

        # Generate output filename based on directory name
        dir_name: str = os.path.basename(os.path.normpath(directory_path))
        filename: str = f"{dir_name}_metadata.json"

        if output_dir is not None:
            # Ensure output directory exists
            os.makedirs(output_dir, exist_ok=True)
            output_path = os.path.join(output_dir, filename)
        else:
            # If no output directory specified, save in the current directory
            output_path = filename

        # Write all metadata to a single JSON file
        with open(output_path, "w", encoding="utf-8") as f:
            json.dump(all_metadata, f, indent=2, ensure_ascii=False)

        self.logger.debug(f"Metadata for {file_count} files saved to {output_path}")
        return True