let selected = []
let currentPath = []

const SIZES = ["", "K", "M", "G", "T"]

function formatSize(bytes) {
    let order = Math.floor(Math.log10(bytes) / 3)
    if (bytes > Math.pow(10, (order + 1) * 3) / 2) {
        order += 1
    }
    let size = bytes / Math.pow(10, order * 3)
    size = Math.round(size * 10) / 10
    const prefix = SIZES[order]
    return `${size}${prefix}B`
}

function makeFolder(meta) {
    const file = document.getElementById("folder-template").cloneNode(true)
    file.querySelector(".name").innerText = meta.filename
    file.querySelector(".children .num").innerText = meta.elements
    file.addEventListener("dblclick", () => {
        currentPath.push(meta.filename)
        navigate()
    })
    return file
}

function makeMedia(meta) {
    const file = document.getElementById("media-template").cloneNode(true)
    file.querySelector(".name").innerText = meta.filename.split(".").slice(0, -1).join(".")
    file.querySelector(".size").innerText = formatSize(meta.size)
    file.querySelector(".ext").innerText = meta.filename.split(".").slice(-1)[0].toUpperCase()
    return file
}

function makeFile(meta) {
    let file
    switch (meta.type) {
        case "folder":
            file = makeFolder(meta)
            break
        case "media":
            file = makeMedia(meta)
            break
        default:
            throw new Error(`Invalid file type '${meta.type}'`)
    }

    file.title = meta.filename
    file.id = null
    file.dataset.path = meta.path
    file.classList.remove("template")
    if (selected.includes(meta.path)) {
        file.classList.add("selected")
    }
    
    file.addEventListener("click", e => {
        if (e.ctrlKey) {
            e.preventDefault()
            toggleSelectFile(meta)
        }
    })
    return file
}

function makeAgent(meta) {
    const agent = document.getElementById("agent-template").cloneNode(true)
    agent.classList.remove("template")
    agent.removeAttribute("id")
    agent.dataset.uuid = meta.uuid
    agent.querySelector(".name").innerText = meta.name
    const input = agent.querySelector("input[type='radio']")
    input.value = meta.uuid
    input.addEventListener("change", () => updateConvertBtn())
    return agent
}

function toggleSelectFile(meta) {
    if (selected.includes(meta.path)) {
        deselectFile(meta.path)
    } else {
        selectFile(meta)
    }
}

function selectFile(meta) {
    selected.push(meta.path)
    document.querySelector(`.file[data-path='${meta.path}']`)?.classList?.add("selected")
    const list = document.getElementById("selected")
    const line = document.getElementById("selected-template").cloneNode(true)
    line.classList.remove("template")
    line.id = null
    line.querySelector(".name").innerText = meta.path
    line.dataset.path = meta.path
    line.querySelector(".deselect").addEventListener("click", () => {
        deselectFile(meta.path)
    })
    list.appendChild(line)

    const continueBtn = document.getElementById("continue")
    continueBtn.disabled = false
}

function deselectFile(path) {
    selected = selected.filter(p => p !== path)
    document.querySelector(`#files .file[data-path='${path}']`)?.classList?.remove("selected")
    document.querySelector(`#selected .file[data-path='${path}']`)?.remove()

    const continueBtn = document.getElementById("continue")
    continueBtn.disabled = selected.length === 0
}

function deselectAll() {
    selected.forEach(path => {
        deselectFile(path)
    })
    showFiles()
}

/**
 * 
 * @param {object[]} files 
 */
function addFiles(files) {
    const emptyMsg = document.getElementById("no-file")
    if (files.length === 0) {
        emptyMsg.classList.add("show")
    } else {
        emptyMsg.classList.remove("show")
    }
    const list = document.getElementById("files")
    const list2 = document.createElement("div")
    list2.innerHTML = ""
    const filenames = files.map(meta => meta.filename)
    // Copy array because sort changes it in place
    Array.from(filenames).sort().forEach(filename => {
        const i = filenames.indexOf(filename)
        const meta = files[i]
        const file = makeFile(meta)
        list2.appendChild(file)
    })
    list.replaceChildren(...list2.children)

    const upBtn = document.getElementById("up")
    if (currentPath.length === 0) {
        upBtn.classList.remove("show")
    } else {
        upBtn.classList.add("show")
    }
}

function navigate() {
    const url = new URL("/api/files/to_convert", window.location.origin)
    url.searchParams.set("f", currentPath.join("/"))

    fetch(url.href).then(res => {
        return res.json()
    }).then(files => {
        addFiles(files)
    })
}

function showFiles() {
    document.getElementById("show-files").classList.add("hidden")
    document.getElementById("continue").classList.remove("hidden")
    document.getElementById("agents-panel").classList.add("hidden")
    document.getElementById("to-convert-panel").classList.remove("hidden")
}

function showAgents() {
    document.getElementById("continue").classList.add("hidden")
    document.getElementById("show-files").classList.remove("hidden")
    document.getElementById("to-convert-panel").classList.add("hidden")
    document.getElementById("agents-panel").classList.remove("hidden")
}

function updateConvertBtn() {
    const agent = document.querySelector("#agents .agent input:checked")
    const convertBtn = document.getElementById("convert")
    convertBtn.disabled = !agent
}

function addAgents(agents) {
    const emptyMsg = document.getElementById("no-agent")
    if (agents.length === 0) {
        emptyMsg.classList.add("show")
    } else {
        emptyMsg.classList.remove("show")
    }
    const selectedAgent = document.querySelector("#agents .agent input:checked")?.parentElement

    const list = document.getElementById("agents")
    const list2 = document.createElement("div")
    list2.innerHTML = ""
    const agentNames = agents.map(meta => meta.name)
    // Copy array because sort changes it in place
    Array.from(agentNames).sort().forEach(name => {
        const i = agentNames.indexOf(name)
        const meta = agents[i]
        const agent = makeAgent(meta)
        if (selectedAgent?.dataset?.uuid === meta.uuid) {
            agent.querySelector("input").checked = true
        }
        list2.appendChild(agent)
    })
    list.replaceChildren(...list2.children)
    updateConvertBtn()
}

window.addEventListener("load", () => {
    document.getElementById("up").addEventListener("dblclick", () => {
        if (currentPath.length !== 0) {
            currentPath = currentPath.slice(0, -1)
            navigate()
        }
    })
    document.getElementById("show-files").addEventListener("click", () => {
        showFiles()
    })
    document.getElementById("deselect-all").addEventListener("click", () => {
        deselectAll()
    })
    document.getElementById("continue").addEventListener("click", () => {
        showAgents()
    })
    navigate()
})